

class CommandGenerator:
    def __init__(self, *,
                 num_agents,
                 num_evals,
                 executable,
                 config_commands=None,
                 ignore_python='auto'):
        """
        Args:
            num_agents: number of agents to run
            num_evals: number of evals to run
            executable: path to .py executable in the pod
            config_commands (arr): additional commands that pass
                to user-defined config.py, after "--"
            ignore_python (Ture, False, auto):
                when True, omit ["python", -u"],
                'auto': True when executable does not end with '.py'
                useful for using excecutables
        """
        self.num_agents = num_agents
        self.num_evals = num_evals
        self.executable = executable
        self.config_commands = config_commands
        if ignore_python == 'auto':
            ignore_python = executable[-3:] != '.py'
        self.ignore_python = ignore_python

    def get_command(self, role):
        command = []
        if not self.ignore_python:
            command += ['python', '-u']
        command += [self.executable]
        command += [role]
        command += ['--']
        if self.config_commands is not None:
            command += self.config_commands
        return ' '.join(command)



import time
import numpy as np
import random
from .base import Replay

class DummyReplay(Replay):
    def __init__(self, *args, sampling_start_size, **kwargs):
        super().__init__(*args, **kwargs)
        self._sampling_start_size = sampling_start_size
        self._memory = {}

    def insert(self, exp_dict):
        print('INSERT', exp_dict['exp_pointer'])
        time.sleep(0.2)
        evicted = []
        key = len(self._memory)
        if key in self._memory:
            evicted.append(self._memory[key])
        self._memory[key] = exp_dict
        if evicted:
            print('INSERT passive evict', evicted[0]['exp_pointer'])
        return evicted

    def sample(self, batch_size, batch_i):
        samps = []
        print('SAMPLE START total memory', len(self._memory), 'batch_i', batch_i)
        assert self._start_sample_condition()
        indices = [random.randint(0, len(self._memory) - 1)
                   for _ in range(batch_size)]
        memkeys = list(self._memory.keys())
        for i in indices:
            time.sleep(.3)
            samps.append(self._memory[memkeys[i]])
        print('SAMPLE DONE')
        return samps

    def evict(self, evict_size):
        evict_keys = list(self._memory.keys())[:evict_size]
        print('EVICT START')
        time.sleep(1)
        exps = []
        for k in evict_keys:
            exps.append(self._memory.pop(k))
        print('EVICT DONE:', [exp['exp_pointer'] for exp in exps],
              [exp['obs_pointers'] for exp in exps])
        return exps

    def start_sample_condition(self):
        return len(self._memory) > self._sampling_start_size



from .wrapper import Wrapper
from surreal.session import Config, extend_config, BASE_SESSION_CONFIG, BASE_LEARNER_CONFIG, ConfigError
from surreal.distributed.exp_sender import ExpSender
from collections import deque
import os
import copy


# https://effectivepython.com/2015/02/02/register-class-existence-with-metaclasses/
class ExpSenderWrapperBase(Wrapper):
    def __init__(self, env, learner_config, session_config):
        """
        Default sender configs are in BASE_SESSION_CONFIG['sender']
        They contain communication level information
        
        Algorithm specific experience generation parameters should live in learner_config
        """
        super().__init__(env)
        # TODO: initialize config in a unified place 
        self.session_config = Config(session_config).extend(BASE_SESSION_CONFIG)
        self.learner_config = Config(learner_config).extend(BASE_LEARNER_CONFIG)
        host = os.environ['SYMPH_COLLECTOR_FRONTEND_HOST']
        port = os.environ['SYMPH_COLLECTOR_FRONTEND_PORT']
        self.sender = ExpSender(
            host=host,
            port=port,
            flush_iteration=self.session_config.sender.flush_iteration,
        )
        

class ExpSenderWrapperSSAR(ExpSenderWrapperBase):
    """
        Sends experience in format
        {   
            'obs': [state, next state]
            'action': action,
            'reward': reward,
            'done': done,
            'info': info
        }
    """
    def __init__(self, env, learner_config, session_config):
        super().__init__(env, learner_config, session_config)
        self._obs = None  # obs of the current time step

    def _reset(self):
        self._obs, info = self.env.reset()
        return self._obs, info

    def _step(self, action):
        obs_next, reward, done, info = self.env.step(action)
        self.send([[self._obs, obs_next], action, reward, done, info])
        self._obs = obs_next
        return obs_next, reward, done, info

    def send(self, data):
        obs_array, action, reward, done, info = data
        hash_dict = {
            'obs': obs_array,
        }
        nonhash_dict = {
            'action': action,
            'reward': reward,
            'done': done,
            'info': info
        }
        self.sender.send(hash_dict, nonhash_dict)

# Naming may need some change here. 
# The unit of experience is in format state state action reward.
# N-step is supported
class ExpSenderWrapperSSARNStepBootstrap(ExpSenderWrapperSSAR):
    """
        Sends observations in format
        {   
            'obs': [state, next_state]
            'action': action,
            'reward': reward,
            'done': done,
            'info': info
        }
        but next_state is n_steps after state and reward is cumulated reward over n_steps
        Used for n_step reward computations.

        Requires:
            @self.learner_config.algo.n_step: number of steps to cumulate over
            @self.learner_config.algo.gamma: discount factor
    """
    def __init__(self, env, learner_config, session_config):
        super().__init__(env, learner_config, session_config)
        self.n_step = self.learner_config.algo.n_step
        self.gamma = self.learner_config.algo.gamma
        self.last_n = deque()

    def _reset(self):
        self._obs, info = self.env.reset()
        self.last_n.clear()
        return self._obs, info

    def _step(self, action):
        obs_next, reward, done, info = self.env.step(action)
        for i, exp_list in enumerate(self.last_n):
            # Update Next Observation and done to be from the final of n_steps and reward to be weighted sum
            exp_list[0][1] = obs_next
            exp_list[2] += pow(self.gamma, self.n_step - i - 1) * reward
            exp_list[3] = done
        self.last_n.append([[self._obs, obs_next], action, reward, done, info])
        if len(self.last_n) == self.n_step:
            self.send(self.last_n.popleft())

        self._obs = obs_next
        return obs_next, reward, done, info

class ExpSenderWrapperMultiStep(ExpSenderWrapperBase):
    """
        Base class for all classes that send experience in format
        {   
            'obs_arr': [state_1, ..., state_n]
            'obs_next': [state_{n + 1}]
            'action_arr': [action_1, ...],
            'reward_arr': [reward_1, ...],
            'done_arr': [done_1, ...],
            'info_arr': [info_1, ...],
            'n_step': n, length of all arrays,
        }
    """
    def send(self, data, obs_next):
        obs_arr, action_arr, reward_arr, done_arr, info_arr = [], [], [], [], []
        hash_dict = {}
        nonhash_dict = {}
        for index, (obs, action, reward, done, info) in enumerate(data):
            # Store observations in a deduplicated way
            obs_arr.append(obs)
            action_arr.append(action)
            reward_arr.append(reward)
            done_arr.append(done)
            info_arr.append(info)

        hash_dict = {
        }
        nonhash_dict = {
            'obs_arr': obs_arr,
            'obs_next': obs_next,
            'action_arr': action_arr,
            'reward_arr': reward_arr,
            'done_arr': done_arr,
            'info_arr': info_arr,
            'n_step': len(data),
        }
        self.sender.send(hash_dict, nonhash_dict)


class ExpSenderWrapperMultiStepMovingWindowWithInfo(ExpSenderWrapperBase):
    """
        Base class for all classes that send experience in format
        {   
            'obs': [state_1, ..., state_n]
            'obs_next': [state_{n + 1}]
            'actions': [action_1, ...],
            'rewards': [reward_1, ...],
            'dones': [done_1, ...],
            'persistent_infos': [infolist_1, ...],
            'onetime_infos': [infos],
            'infos': [info_1, ...],
            'n_step': n
        }

        Note: distinction between persistent infos and onetime infos is subtle.
        Persistent info is collected for each step of rollout (i.e. prob distrib)
        Onetime info is collected once for the first state in partial trajectory,
            i.e. LSTM hidden/cell states

        Requires:
            @self.learner_config.algo.n_step: n, number of steps per experience
            @self.learner_config.algo.stride: after sending experience [state_i, ...]
            the next experience is [state_{i + stride}]
    """
    def __init__(self, env, learner_config, session_config):
        '''
        Consturctor for ExpSenderWrapperMultiStepMovingWindowWithInfo class
        Important Attributes:
            _ob: holds the state at current timestep
            n_step: maximum number of previous states to keep
            stride: stride for moving window
            last_n: queue of max size n_step to hold previous states
            env: environment to interact with
            sender: distributed experience sender. communicate with local and
                    remote replay buffer
        '''
        super().__init__(env, learner_config, session_config)
        self._ob = None  # obs of the current time step
        self.n_step = self.learner_config.algo.n_step
        self.stride = self.learner_config.algo.stride # Stride for moving window
        if self.stride < 1:
            raise ConfigError('stride {} for experience generation cannot be less than 1'.format(self.learner_config.algo.stride))
        self.last_n = deque()

    def _reset(self):
        '''
            Reset current states at the beginning of an episode
            Note: deepcopy is required to prevent Mujoco changing the states
                  under the hood
        '''
        obs, info = self.env.reset()
        self._ob = copy.deepcopy(obs)
        self.last_n.clear()
        return self._ob, info

    def _step(self, action):
        '''
            Step function interact with underlying environment. Adds this state
            to maintained queue. If stide and length criteria is met, experiences
            are sent to distributed buffer
            Args: 
                action: numpy array of shape (1 x action_dim) that denotes
                        action generated by actor
        '''
        action_choice, action_info = action
        obs_next, reward, done, info = self.env.step(action_choice)
        self.last_n.append([self._ob, action_choice, reward, done, action_info[0], action_info[1], info])

        if len(self.last_n) == self.n_step:
            self.send(self.last_n, obs_next)
            for i in range(self.stride):
                if len(self.last_n) > 0:
                    self.last_n.popleft()
        self._ob = copy.deepcopy(obs_next)
        return obs_next, reward, done, info

    def send(self, data, obs_next):
        '''
            Sends the current experiences to distributed replay buffer
            Args: 
                data: list of sub-trajectories collected
                obs_next: next observation
        '''
        obs, actions, rewards, dones, persistent_infos, infos = [], [], [], [], [], []
        hash_dict = {}
        nonhash_dict = {}
        onetime_infos = None
        for index, (ob, action, reward, done, onetime_info, persistent_info, info) in enumerate(data):
            # Store observations in a deduplicated way
            obs.append(ob)
            actions.append(action)
            rewards.append(reward)
            dones.append(done)
            infos.append(info)
            persistent_infos.append(persistent_info)
            if onetime_infos == None: onetime_infos = onetime_info

        hash_dict = {
            'obs': obs,
            'obs_next': obs_next
        }
        nonhash_dict = {
            'actions': actions,
            'onetime_infos' : onetime_infos,
            'persistent_infos': persistent_infos,
            'rewards': rewards,
            'dones': dones,
            'infos': infos,
            'n_step': len(data),
        }            
        self.sender.send(hash_dict, nonhash_dict)


class ExpSenderWrapperMultiStepMovingWindow(ExpSenderWrapperMultiStep):
    """
        Base class for all classes that send experience in format
        {   
            'obs_arr': [state_1, ..., state_n]
            'obs_next': [state_{n + 1}]
            'action_arr': [action_1, ...],
            'reward_arr': [reward_1, ...],
            'done_arr': [done_1, ...],
            'info_arr': [info_1, ...],
            'n_step': n
        }

        Requires:
            @self.learner_config.algo.n_step: n, number of steps per experience
            @self.learner_config.algo.stride: after sending experience [state_i, ...]
            the next experience is [state_{i + stride}]
    """
    def __init__(self, env, learner_config, session_config):
        super().__init__(env, learner_config, session_config)
        self._obs = None  # obs of the current time step
        self.n_step = self.learner_config.algo.n_step
        self.stride = self.learner_config.algo.stride # Stride for moving window
        if self.stride < 1:
            raise ConfigError('stride {} for experience generation cannot be less than 1'.format(self.learner_config.algo.stride))
        self.last_n = deque()

    def _reset(self):
        self._obs, info = self.env.reset()
        self.last_n.clear()
        return self._obs, info

    def _step(self, action):
        obs_next, reward, done, info = self.env.step(action)
        self.last_n.append([self._obs, action, reward, done, info])
        if len(self.last_n) == self.n_step:
            self.send(self.last_n, obs_next)
            for i in range(self.stride):
                if len(self.last_n) > 0:
                    self.last_n.popleft()
        self._obs = obs_next
        return obs_next, reward, done, info



import os
from surreal.utils.filesys import f_expand


def get_config_file():
    """
    Get the path to
    """
    path = '~/.surreal.yml'
    if 'SURREAL_CONFIG_PATH' in os.environ:
        path = os.environ['SURREAL_CONFIG_PATH']
    return f_expand(path)



# Adapted from https://github.com/openai/baselines

import numpy as np
import random

from segment_tree import SumSegmentTree, MinSegmentTree


class PrioritizedReplay(Replay):
    
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config):
        """
        Create Prioritized Replay buffer.
        :param size: int
            Max number of transitions to store in the buffer. When the buffer
            overflows the old memories are dropped.
        :param alpha: float
            how much prioritization is used
            (0 - no prioritization, 1 - full prioritization)
        """
        super(PrioritizedReplayBuffer, self).__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config
        )

        self._alpha = self.replay_config.alpha
        assert self._alpha > 0

        self._memory = []
        self.memory_size = self.replay_config.memory_size
        it_capacity = 1
        while it_capacity < self.memory_size:
            it_capacity *= 2

        self._it_sum = SumSegmentTree(it_capacity)
        self._it_min = MinSegmentTree(it_capacity)
        self._max_priority = 1.0

    def default_config(self):
        conf = super().default_config()
        conf.update({
            'memory_size': '_int_',
            'sampling_start_size': '_int_',
            'alpha': '_float_',
        })
        return conf

    def insert(self, exp_dict):
        """
        Adds experience to the replay buffer as usual, but also
        intiialize the priority of the new experience.
        """
        with self.insert_time.time():
            if self._next_idx >= len(self._memory):
                self._memory.append(exp_dict)
            else:
                self._memory[self._next_idx] = exp_dict
            self._next_idx = (self._next_idx + 1) % self.memory_size

            idx = self._next_idx
            self._it_sum[idx] = self._max_priority ** self._alpha
            self._it_min[idx] = self._max_priority ** self._alpha

    def sample(self, batch_size, beta=0):
        """
        WARNING: This function does not make deep copies of the tuple experiences.
                 This means that if any objects in the experiences are modified,
                 the contents of the replay buffer memory will also be modified,
                 so be careful!!!
        Sample a batch of experiences, along with their importance weights, and the 
        indices of the sampled experiences in the buffer.
        :param batch_size: int
            How many transitions to sample.
        :param beta: float
            To what degree to use importance weights
            (0 - no corrections, 1 - full correction)
        :return experience_batch: List
            List of tuples, length batch_size, corresponding to the experiences sampled.
        :return weights: np.array
            Array of shape (batch_size,) and dtype np.float32
            denoting importance weight of each sampled transition
        :return indices: np.array
            Array of shape (batch_size,) and dtype np.int32
            indices in buffer of sampled experiences
        """
        with self.sample_time.time():
            assert beta >= 0

            # sample the experiences proportional to their priorities
            indices = self._sample_proportional(batch_size)
            response = [self._storage[idx] for idx in indices]

            # compute importance weights for the experiences to correct for distribution shift
            weights = []
            p_min = self._it_min.min() / self._it_sum.sum()
            max_weight = (p_min * len(self._storage)) ** (-beta)

            for idx in indices:
                p_sample = self._it_sum[idx] / self._it_sum.sum()
                weight = (p_sample * len(self._storage)) ** (-beta)
                weights.append(weight / max_weight)
            weights = np.array(weights)

        # return response, weights, indices
        return response

    def _sample_proportional(self, batch_size):
        """
        This is a helper function to sample expriences with probabilities 
        proportional to their priorities.
        Returns a list of indices.
        """
        res = []
        for _ in range(batch_size):
            mass = random.random() * self._it_sum.sum(0, len(self._storage) - 1)
            idx = self._it_sum.find_prefixsum_idx(mass)
            res.append(idx)
        return res

    def update_priorities(self, indices, priorities):
        """
        Update priorities of sampled transitions.
        sets priority of transition at index indices[i] in buffer
        to priorities[i].
        :param indices: [int]
            List of indices of sampled transitions
        :param priorities: [float]
            List of updated priorities corresponding to
            transitions at the sampled indices denoted by
            variable `indices`.
        """
        assert len(indices) == len(priorities)
        for idx, priority in zip(indices, priorities):
            assert priority > 0
            assert 0 <= idx < len(self._storage)
            self._it_sum[idx] = priority ** self._alpha
            self._it_min[idx] = priority ** self._alpha

            self._max_priority = max(self._max_priority, priority)

    def evict(self):
        raise NotImplementedError # TODO
        # if evict_size > len(self._memory):
        #     evicted = self._memory
        #     self._memory = []
        #     self._next_idx = 0
        #     return evicted
        # forward_space = len(self._memory) - self._next_idx
        # if evict_size < forward_space:
        #     evicted = self._memory[self._next_idx:self._next_idx+evict_size]
        #     del self._memory[self._next_idx:self._next_idx+evict_size]
        # else:
        #     evicted = self._memory[self._next_idx:]
        #     evict_from_left = evict_size - forward_space
        #     evicted += self._memory[:evict_from_left]
        #     del self._memory[self._next_idx:]
        #     del self._memory[:evict_from_left]
        #     self._next_idx -= evict_from_left
        # assert len(evicted) == evict_size
        # return evicted

    def start_sample_condition(self):
        return len(self) > self.replay_config.sampling_start_size

    def __len__(self):
        return len(self._memory)



import random
from .base import Replay
import surreal.utils as U


class UniformReplay(Replay):
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 index=0):
        """
        Args:
          memory_size: Max number of experience to store in the buffer.
            When the buffer overflows the old memories are dropped.
          sampling_start_size: min number of exp above which we will start sampling
        """
        super().__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            index=index
        )
        self._memory = []
        self.memory_size = self.learner_config.replay.memory_size
        self._next_idx = 0

    # def default_config(self):
    #     conf = super().default_config()
    #     conf.update({
    #         'memory_size': '_int_',
    #         'sampling_start_size': '_int_'
    #     })
    #     return conf

    def insert(self, exp_dict):
        if self._next_idx >= len(self._memory):
            self._memory.append(exp_dict)
        else:
            self._memory[self._next_idx] = exp_dict
        self._next_idx = (self._next_idx + 1) % self.memory_size

    def sample(self, batch_size):    
        indices = [random.randint(0, len(self._memory) - 1)
                   for _ in range(batch_size)]
        response = [self._memory[i] for i in indices]
        return response

    def evict(self):
        raise NotImplementedError  # TODO
        if evict_size > len(self._memory):
            evicted = self._memory
            self._memory = []
            self._next_idx = 0
            return evicted
        forward_space = len(self._memory) - self._next_idx
        if evict_size < forward_space:
            evicted = self._memory[self._next_idx:self._next_idx+evict_size]
            del self._memory[self._next_idx:self._next_idx+evict_size]
        else:
            evicted = self._memory[self._next_idx:]
            evict_from_left = evict_size - forward_space
            evicted += self._memory[:evict_from_left]
            del self._memory[self._next_idx:]
            del self._memory[:evict_from_left]
            self._next_idx -= evict_from_left
        assert len(evicted) == evict_size
        return evicted

    def start_sample_condition(self):
        return len(self) > self.learner_config.replay.sampling_start_size

    def __len__(self):
        return len(self._memory)



import time
import json
import numpy as np
from tabulate import tabulate
from collections import OrderedDict
from surreal.session import (PeriodicTracker, get_tensorplex_client)
import surreal.utils as U
from .wrapper import Wrapper


class EpisodeMonitor(Wrapper):
    """
    Access the public properties to get episode history:
    - episode_rewards
    - episode_steps
    - total_steps
    - episode_durations (in seconds)
    - num_episodes == len(episode_rewards)
    """
    def __init__(self, env):
        super().__init__(env)
        self._tstart_ep0 = time.time()
        self._tstart_current_ep = None
        self._rewards_current_ep = None
        self.episode_rewards = []
        self.episode_steps = []
        self.episode_durations = []
        self.total_steps = 0

    def _reset(self, **kwargs):
        self._rewards_current_ep = []
        self._tstart_current_ep = time.time()
        return self.env.reset(**kwargs)

    def _step(self, action):
        ob, rew, done, info = self.env.step(action)
        self._rewards_current_ep.append(rew)
        if done:
            eprew = round(sum(self._rewards_current_ep), 6)
            epsteps = len(self._rewards_current_ep)
            eptime = round(time.time() - self._tstart_current_ep, 6)
            epinfo = {
                "reward": eprew,
                "steps": epsteps,
                "duration": eptime,
                "total_elapsed": round(time.time() - self._tstart_ep0, 6),
            }
            self.episode_rewards.append(eprew)
            self.episode_steps.append(epsteps)
            self.episode_durations.append(eptime)
            info['episode'] = epinfo
        self.total_steps += 1
        return ob, rew, done, info

    @property
    def num_episodes(self):
        return len(self.episode_rewards)

    def step_per_sec(self, average_episodes):
        """
        Speedometer, step per second

        Args:
            average_episodes: average over the past N episodes
        """
        assert average_episodes > 0
        return (sum(self.episode_steps[-average_episodes:])
                / (sum(self.episode_durations[-average_episodes:]) + 1e-7))


class ConsoleMonitor(EpisodeMonitor):
    def __init__(self, env,
                 update_interval=10,
                 average_over=10,
                 extra_rows=None):
        """
        Args:
            update_interval: print every N episodes
            average_over: average rewards/speed over the last N episodes
            extra_rows: an OrderedDict {'row caption': function(total_steps, num_episodes)}
                to generate extra rows to the printed table.
        """
        super().__init__(env)
        self._periodic = PeriodicTracker(update_interval)
        self._avg = average_over
        if extra_rows is None:
            self._extra_rows = OrderedDict()
        else:
            assert isinstance(extra_rows, OrderedDict), \
                'extra_rows spec {"row caption": function(total_steps, ' \
                'num_episodes)} must be an OrderedDict'
            self._extra_rows = extra_rows

    def _step(self, action):
        ob, r, done, info = super()._step(action)
        if done and self._periodic.track_increment():
            info_table = []
            avg_reward = U.mean(self.episode_rewards[-self._avg:])
            info_table.append(['Last {} rewards'.format(self._avg),
                               U.fformat(avg_reward, 3)])
            avg_speed = self.step_per_sec(self._avg)
            info_table.append(['Speed iter/s',
                               U.fformat(avg_speed, 1)])
            info_table.append(['Total steps', self.total_steps])
            info_table.append(['Episodes', self.num_episodes])
            for row_caption, row_func in self._extra_rows.items():
                row_value = row_func(self.total_steps, self.num_episodes)
                info_table.append([row_caption, str(row_value)])
            # `fancy_grid` doesn't work in terminal that doesn't display unicode
            print(tabulate(info_table, tablefmt='simple', numalign='left'))
        return ob, r, done, info


class TrainingTensorplexMonitor(EpisodeMonitor):
    def __init__(self, env,
                 agent_id,
                 session_config,
                 separate_plots=True):
        """
        Display "reward" and "step_per_s" curves on Tensorboard

        Args:
            env:
            agent_id: int.
            session_config: to construct AgentTensorplex
            - interval: log to Tensorplex every N episodes.
            - average_episodes: average rewards/speed over the last N episodes
            separate_plots: True to put reward plot in a separate section on
                Tensorboard, False to put all plots together
        """
        super().__init__(env)
        U.assert_type(agent_id, int)
        self.tensorplex = get_tensorplex_client(
            '{}/{}'.format('agent', agent_id),
            session_config
        )
        interval = session_config['tensorplex']['update_schedule']['training_env']
        self._periodic = PeriodicTracker(interval)
        self._avg = interval
        self._separate_plots = separate_plots

    def _get_tag(self, tag):
        if self._separate_plots:
            return ':' + tag  # see Tensorplex tag semantics
        else:
            return tag

    def _step(self, action):
        ob, r, done, info = super()._step(action)
        if done and self._periodic.track_increment():
            scalar_values = {
                self._get_tag('reward'):
                    U.mean(self.episode_rewards[-self._avg:]),
                'step_per_s':
                    self.step_per_sec(self._avg),
            }
            self.tensorplex.add_scalars(
                scalar_values,
                global_step=self.num_episodes
            )
        return ob, r, done, info


class EvalTensorplexMonitor(EpisodeMonitor):
    def __init__(self, env,
                 eval_id,
                 fetch_parameter,
                 session_config,
                 separate_plots=False):
        """
        Display "reward" and "step_per_s" curves on Tensorboard

        Args:
            env:
            eval_id:
            fetch_parameter: lambda function that pulls from parameter server
            session_config: to construct AgentTensorplex
            - interval: log to Tensorplex every N episodes.
            - average_episodes: average rewards/speed over the last N episodes
            separate_plots: True to put reward plot in a separate section on
                Tensorboard, False to put all plots together
        """
        super().__init__(env)
        self.tensorplex = get_tensorplex_client(
            '{}/{}'.format('eval', eval_id),
            session_config
        )
        interval = session_config['tensorplex']['update_schedule']['eval_env']
        self._periodic = PeriodicTracker(interval)
        self._avg = interval
        self._separate_plots = separate_plots
        self._throttle_sleep = \
            session_config['tensorplex']['update_schedule']['eval_env_sleep']
        self._fetch_parameter = fetch_parameter
        self._fetch_parameter()  # if this eval is late to the party

    def _get_tag(self, tag):
        if self._separate_plots:
            return ':' + tag  # see Tensorplex tag semantics
        else:
            return tag

    def _step(self, action):
        ob, r, done, info = super()._step(action)
        if done and self._periodic.track_increment():
            scalar_values = {
                self._get_tag('reward'):
                    U.mean(self.episode_rewards[-self._avg:]),
                'step_per_s':
                    self.step_per_sec(self._avg),
            }
            self.tensorplex.add_scalars(
                scalar_values,
                global_step=self.num_episodes
            )
            time.sleep(self._throttle_sleep)
            self._fetch_parameter()
        return ob, r, done, info



import os
from surreal.env.video_env import VideoWrapper
import surreal.utils as U
from .wrapper import (
    GymAdapter,
    FrameStackWrapper,
    GrayscaleWrapper,
    TransposeWrapper,
    FilterWrapper,
    ObservationConcatenationWrapper,
    RobosuiteWrapper
    )


def make_env(env_config, mode=None):
    """
    Makes an environment and populates related fields in env_config
    return env, env_config
    Args:
        overrides(str): point to the override in env_config to 
                        provide differnt kwargs to the initializer of the environment
    """
    env_name = env_config.env_name
    env_category, env_name = env_name.split(':')
    if mode == 'eval' and 'eval_mode' in env_config:
        for k, v in env_config.eval_mode.items():
            env_config[k] = v
    if env_category == 'gym':
        env, env_config = make_gym(env_name, env_config)
    elif env_category == 'robosuite':
        env, env_config = make_robosuite(env_name, env_config)
    elif env_category == 'dm_control':
        env, env_config = make_dm_control(env_name, env_config)
    else:
        raise ValueError('Unknown environment category: {}'.format(env_category))
    return env, env_config


def make_gym(env_name, env_config):
    import gym
    env = gym.make(env_name)
    env = GymAdapter(env, env_config)
    env_config.action_spec = env.action_spec()
    env_config.obs_spec = env.observation_spec()
    return env, env_config


def make_robosuite(env_name, env_config):
    import robosuite

    env = robosuite.make(
        env_name,
        has_renderer=False,
        ignore_done=True,
        use_camera_obs=env_config.pixel_input,
        camera_height=84,
        camera_width=84,
        render_collision_mesh=False,
        render_visual_mesh=True,
        camera_name='agentview',
        use_object_obs=(not env_config.pixel_input),
        camera_depth=env_config.use_depth,
        reward_shaping=True,
        # demo_config=env_config.demonstration,
    )
    env = RobosuiteWrapper(env, env_config)
    env = FilterWrapper(env, env_config)
    env = ObservationConcatenationWrapper(env)
    if env_config.pixel_input:
        env = TransposeWrapper(env)
        if env_config.use_grayscale:
            env = GrayscaleWrapper(env)
        if env_config.frame_stacks:
            env = FrameStackWrapper(env, env_config)
    env_config.action_spec = env.action_spec()
    env_config.obs_spec = env.observation_spec()
    return env, env_config


def make_dm_control(env_name, env_config):
    from dm_control import suite
    from dm_control.suite.wrappers import pixels
    from .dm_wrapper import DMControlAdapter, DMControlDummyWrapper
    pixel_input = env_config.pixel_input
    domain_name, task_name = env_name.split('-')
    env = suite.load(domain_name=domain_name, task_name=task_name)
    if pixel_input:
        if os.getenv('DISABLE_MUJOCO_RENDERING'):
            # We are asking for rendering on a pod that cannot support rendering, 
            # This happens in GPU based learners when we only want to create the environment
            # to see the dimensions.
            # So we will add a dummy environment
            # TODO: add a dummy wrapper that only contains the correct specs
            env = DMControlDummyWrapper(env) #...
        else:
            env = pixels.Wrapper(env, render_kwargs={'height': 84, 'width': 84, 'camera_id': 0})
    # TODO: what to do with reward visualization
    # Reward visualization should only be done in the eval agent
    # env = suite.load(domain_name=domain_name, task_name=task_name, visualize_reward=record_video)
    env = DMControlAdapter(env, pixel_input)
    env = FilterWrapper(env, env_config)
    env = ObservationConcatenationWrapper(env)
    if pixel_input:
        env = TransposeWrapper(env)
        env = GrayscaleWrapper(env)
        if env_config.frame_stacks > 1:
            env = FrameStackWrapper(env, env_config)
    env_config.action_spec = env.action_spec()
    env_config.obs_spec = env.observation_spec()
    return env, env_config



"""
Template class for all learners
"""
import os
import threading
import queue
import time
import numpy as np
from pathlib import Path
from benedict import BeneDict
import surreal.utils as U
from surreal.session import (
    TimeThrottledTensorplex,
    get_loggerplex_client,
    get_tensorplex_client,
    Config
)
from surreal.distributed import ParameterPublisher, LearnerDataPrefetcher


class Learner(metaclass=U.AutoInitializeMeta):
    """
        Important: When extending this class, make sure to follow the init
            method signature so that orchestrating functions can properly
            initialize the learner.
    """
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config):
        """
        Initializes the learner instance

        Args:
            learner_config, env_config, session_config: configs that define
                an experiment
        """
        self.learner_config = learner_config
        self.env_config = env_config
        self.session_config = session_config
        self.current_iter = 0

        self._setup_logging()
        self._setup_checkpoint()

    def learn(self, batch_exp):
        """
        Abstract method runs one step of learning

        Args:
            batch_exp: batched experience, format is a list of whatever
                experience sender wrapper returns

        Returns:
            td_error or other values for prioritized replay
        """
        raise NotImplementedError

    def module_dict(self):
        """
            Dict of modules to be broadcasted to the parameter server.
            MUST be consistent with the agent's `module_dict()`
        """
        raise NotImplementedError

    def save(self, file_path):
        """
            Saves checkpoint to disk

        Args:
            file_path: locatioin to save
        """
        raise NotImplementedError

    def checkpoint_attributes(self):
        """
            This function defines what attributes should be serialized
            when we are saving a checkpoint.

            See implementations in DDPGLearner and PPOLearner for examples

        Returns:
            list of attributes to be tracked by checkpoint
        """
        return []

    ######
    # Internal, including Communication, etc.
    ######
    def _setup_publish(self):
        min_publish_interval = \
            self.learner_config.parameter_publish.min_publish_interval
        self._ps_publish_tracker = U.TimedTracker(min_publish_interval)

        ps_publish_port = os.environ['SYMPH_PARAMETER_PUBLISH_PORT']
        self._ps_publisher = ParameterPublisher(
            port=ps_publish_port,
            module_dict=self.module_dict()
            # This must happen after subclass __init__
        )

    def _setup_prefetching(self):
        batch_size = self.learner_config.replay.batch_size
        self._prefetch_queue = LearnerDataPrefetcher(
            session_config=self.session_config,
            batch_size=batch_size,
            worker_preprocess=self._prefetcher_preprocess,
            main_preprocess=self.preprocess
        )
        self._prefetch_queue.start()

        # self._preprocess_prefetch_queue = queue.Queue(maxsize=2)
        # self._preprocess_thread = threading.Thread(
        #     target=self._preprocess_batch)
        # self._preprocess_thread.start()

    def _initialize(self):
        """
            For AutoInitializeMeta interface
        """
        self._setup_publish()
        self._setup_prefetching()
        # Logging should only start here so that all components are
        # properly initialized
        self._tensorplex_thread.start()

    ######
    # Parameter publish
    ######
    def should_publish_parameter(self):
        return self._ps_publish_tracker.track_increment()

    def publish_parameter(self, iteration, message=''):
        """
        Learner publishes latest parameters to the parameter server.

        Args:
            iteration: the current number of learning iterations
            message: optional message, must be pickleable.
        """
        self._ps_publisher.publish(iteration, message=message)

    ######
    # Getting data
    ######
    def fetch_batch(self):
        return self._prefetch_queue.get()

    def fetch_iterator(self):
        while True:
            yield self.fetch_batch()

    ######
    # Logging
    ######
    def _setup_logging(self):
        self.learn_timer = U.TimeRecorder()
        self.iter_timer = U.TimeRecorder()
        self.publish_timer = U.TimeRecorder()

        self.init_time = time.time()
        self.current_iter = 0

        self.last_time = self.init_time
        self.last_time_2 = self.init_time
        self.last_iter = 0

        self.log = get_loggerplex_client('learner', self.session_config)
        self.tensorplex = self._get_tensorplex('learner/learner')

        self._tensorplex_thread = U.PeriodicWakeUpWorker(
            target=self.generate_tensorplex_report)

    def _get_tensorplex(self, name):
        """
            Get the periodic tensorplex object
        Args:
            @name: The name of the collection of metrics
        """
        tp = get_tensorplex_client(
            name,
            self.session_config
        )
        update_schedule = self.session_config.tensorplex.update_schedule
        periodic_tp = TimeThrottledTensorplex(
            tensorplex=tp,
            min_update_interval=update_schedule.learner_min_update_interval,
        )
        return periodic_tp

    # def start_tensorplex_thread(self):
    #     if self._tensorplex_thread is not None:
    #         raise RuntimeError('tensorplex thread already running')
    #     self._tensorplex_thread = U.PeriodicWakeUpWorker(target=self.generate_tensorplex_report)
    #     self._tensorplex_thread.start()
    #     return self._tensorplex_thread

    def generate_tensorplex_report(self):
        """
            Adds core and system level tensorplex stats
        """
        cur_time = time.time()
        current_iter = self.current_iter

        iter_elapsed = current_iter - self.last_iter
        self.last_iter = current_iter

        time_elapsed = cur_time - self.last_time
        self.last_time = cur_time

        core_metrics = {}
        system_metrics = {}

        learn_time = self.learn_timer.avg + 1e-6
        fetch_timer = self._prefetch_queue.timer
        fetch_time = fetch_timer.avg + 1e-6
        iter_time = self.iter_timer.avg + 1e-6
        publish_time = self.publish_timer.avg + 1e-6
        # Time it takes to learn from a batch
        core_metrics['learn_time_s'] = learn_time
        # Time it takes to fetch a batch
        core_metrics['fetch_time_s'] = fetch_time
        # Time it takes to publish parameters
        core_metrics['publish_time_s'] = publish_time
        # Time it takes to complete one full iteration
        core_metrics['iter_time_s'] = iter_time

        iter_per_s = iter_elapsed / time_elapsed
        # Number of iterations per second
        system_metrics['iter_per_s'] = iter_per_s
        # Number of experience bathces processed per second
        system_metrics['exp_per_s'] = iter_per_s * \
            self.learner_config.replay.batch_size
        # Percent of time spent on learning
        system_metrics['compute_load_percent'] = min(
            learn_time / iter_time * 100, 100)
        # Percent of time spent on IO
        system_metrics['io_fetch_experience_load_percent'] = min(
            fetch_time / iter_time * 100, 100)
        # Percent of time spent on publishing
        system_metrics['io_publish_load_percent'] = min(
            publish_time / iter_time * 100, 100)

        all_metrics = {}
        for k in core_metrics:
            all_metrics['.core/' + k] = core_metrics[k]
        for k in system_metrics:
            all_metrics['.system/' + k] = system_metrics[k]

        # These are system metrics,
        # they don't add to counter or trigger updates
        self.tensorplex.add_scalars(all_metrics)

    ######
    # Checkpoint
    ######
    def _setup_checkpoint(self):
        # Setup saving
        tracked_attrs = self.checkpoint_attributes()
        assert U.is_sequence(tracked_attrs), \
            'checkpoint_attributes must return a list of string attr names'
        self._periodic_checkpoint = U.PeriodicCheckpoint(
            U.f_join(self.session_config.folder, 'checkpoint'),
            name='learner',
            period=self.session_config.checkpoint.learner.periodic,
            min_interval=self.session_config.checkpoint.learner.min_interval,
            tracked_obj=self,
            tracked_attrs=tracked_attrs,
            keep_history=self.session_config.checkpoint.learner.keep_history,
            keep_best=self.session_config.checkpoint.learner.keep_best,
            # TODO figure out how to add score to learner
        )

        # Load when instructed by config
        # restore_checkpoint should be called _after_ subclass __init__
        # that's why we put it in _initialize()
        if self.session_config.checkpoint.restore:
            self.restore_checkpoint()

    def periodic_checkpoint(self, global_steps, score=None, **info):
        """
        Will only save at the end of each period

        Args:
            global_steps: the number of iterations
            score: the evaluation score for saving the best parameters.
                Currently Not supported!!!
                None when session_config.checkpoint.keep_best=False
            **info: other metadata to save in checkpoint

        Returns:
            saved(bool): whether save() is actually called or not
        """
        return self._periodic_checkpoint.save(
            score=score,
            global_steps=global_steps,
            reload_metadata=False,
            **info,
        )

    def restore_checkpoint(self):
        SC = self.session_config
        restore_folder = SC.checkpoint.restore_folder
        if (restore_folder and
                U.f_last_part_in_path(restore_folder) != 'checkpoint'):
            # automatically append 'checkpoint' subfolder
            restore_folder = U.f_join(restore_folder, 'checkpoint')
        restored = self._periodic_checkpoint.restore(
            target=SC.checkpoint.learner.restore_target,
            mode=SC.checkpoint.learner.mode,
            reload_metadata=True,
            check_ckpt_exists=True,  # TODO set to False unless debugging
            restore_folder=restore_folder,
        )
        if restored:
            self.log.info('successfully restored from checkpoint', restored)

    ######
    # Batch Prefetch
    ######
    def preprocess(self, batch):
        '''
        Perform algorithm-specific preprocessing tasks on a given batch,
        overridden in subclasses
        This operation occurs asynchronously to the learner main loop,
        so if training on gpu, any cpu-bound or high
        latency tasks can be done here.
        For example, ddpg converts relevant variables onto gpu
        '''
        return batch

    def _prefetcher_preprocess(self, batch):
        """

        This function processes the list of experience retrieved from replay
        It happens in a different process from learner's main process

        Args:
            batch: A list of experience
                i.e. a list of packets that agents send to replay

        Returns:
            Data for learner main process
        """
        return batch

    ######
    # Main Loop
    # Override to completely change learner behavior
    ######
    def main(self):
        """
            Main function that defines learner process
        """
        self.main_setup()
        while True:
            self.main_loop()

    def main_setup(self):
        """
            Setup before constant looping
        """
        self.save_config()
        self.iter_timer.start()
        self.publish_parameter(0, message='batch '+str(0))

    def main_loop(self):
        """
            One loop of learner, runs one learn operation of learner
        """
        data = self._prefetch_queue.get()
        with self.learn_timer.time():
            self.learn(data)
        if self.should_publish_parameter():
            with self.publish_timer.time():
                self.publish_parameter(self.current_iter,
                                       message='batch '+str(self.current_iter))
        self.iter_timer.lap()
        self.current_iter += 1

    def save_config(self):
        """
            Save config into a yaml file in root experiment directory
        """
        folder = Path(self.session_config.folder)
        folder.mkdir(exist_ok=True, parents=True)
        config = Config(
            learner_config=self.learner_config,
            env_config=self.env_config,
            session_config=self.session_config
        )
        config.dump_file(str(folder / 'config.yml'))



from .base import Env, ActionType, ObsType
import numpy as np
import surreal.utils as U
import collections
from collections import deque
from operator import mul
import functools
import sys
import gym


class SpecFormat(U.StringEnum):
    SURREAL_CLASSIC = ()
    DM_CONTROL = ()
    MUJOCOMANIP = ()


class Wrapper(Env):
    # Clear metadata so by default we don't override any keys.
    metadata = {}
    # Make sure self.env is always defined, even if things break early.
    env = None

    def __init__(self, env):
        self.env = env
        # Merge with the base metadata
        metadata = self.metadata
        self.metadata = self.env.metadata.copy()
        self.metadata.update(metadata)
        self._ensure_no_double_wrap()
        self._obsspec = None
        # self.obs_spec = env.obs_spec
        # self.action_spec = env.action_spec

    @classmethod
    def class_name(cls):
        return cls.__name__

    def _ensure_no_double_wrap(self):
        env = self.env
        while True:
            if isinstance(env, Wrapper):
                if env.class_name() == self.class_name():
                    raise RuntimeError(
                        "Attempted to double wrap with Wrapper: {}"
                        .format(self.__class__.__name__)
                    )
                env = env.env
            else:
                break

    def step(self, action):
        obs, reward, done, info = super().step(action)
        return obs, reward, done, info

    def _step(self, action):
        return self.env.step(action)

    def _reset(self):
        obs, info = self.env.reset()
        self._assert_conforms_to_spec(obs)
        return obs

    def _render(self, *args, **kwargs):
        return self.env.render(*args, **kwargs)

    def _close(self):
        return self.env.close()

    def _assert_conforms_to_spec(self, obs):
        if not self._obsspec:
            self._obsspec = self.observation_spec()
        for modality in self._obsspec:
            for key in self._obsspec[modality]:
                assert self._obsspec[modality][key] == obs[modality][key].shape

    def __str__(self):
        return '<{}{}>'.format(type(self).__name__, self.env)

    def __repr__(self):
        return str(self)

    def action_spec(self):
        return self.env.action_spec()

    def observation_spec(self):
        return self.env.observation_spec()

    @property
    def unwrapped(self):
        return self.env.unwrapped


class ObsWrapper(Wrapper):
    def _reset(self):
        observation, info = self.env.reset()
        return self._observation(observation), info

    def _step(self, action):
        observation, reward, done, info = self.env.step(action)
        return self.observation(observation), reward, done, info

    def observation(self, observation):
        return self._observation(observation)

    def _observation(self, observation):
        raise NotImplementedError


class RewardWrapper(Wrapper):
    def _step(self, action):
        observation, reward, done, info = self.env.step(action)
        return observation, self.reward(reward), done, info

    def reward(self, reward):
        return self._reward(reward)

    def _reward(self, reward):
        raise NotImplementedError


class ActionWrapper(Wrapper):
    def _step(self, action):
        action = self.action(action)
        return self.env.step(action)

    def action(self, action):
        return self._action(action)

    def _action(self, action):
        raise NotImplementedError

    def reverse_action(self, action):
        return self._reverse_action(action)

    def _reverse_action(self, action):
        raise NotImplementedError

class MaxStepWrapper(Wrapper):
    """
        Simple wrapper to limit maximum steps of an environment
    """
    def __init__(self, env, max_steps):
        super().__init__(env)
        if max_steps <= 0:
            raise ValueError('MaxStepWrapper received max_steps')
        self.max_steps = max_steps
        self.current_step = 0

    def _reset(self):
        self.current_step = 0
        return self.env.reset()

    def _step(self, action):
        self.current_step += 1
        observation, reward, done, info = self.env.step(action)
        if self.current_step >= self.max_steps:
            done = True
        return observation, reward, done, info

# putting import inside to allow difference in dependency
class GymAdapter(Wrapper):
    def __init__(self, env, env_config):
        super().__init__(env)
        assert not env_config.pixel_input, "Pixel input training not supported with OpenAI Gym"
        assert isinstance(env, gym.Env)
        self.env = env

    def _add_modality(self, obs):
        obs = {
            'flat_inputs': obs
        }
        return collections.OrderedDict([('low_dim', obs)])

    def _reset(self):
        obs = self.env.reset()
        return self._add_modality(obs), {}

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        obs = self._add_modality(obs)
        return obs, reward, done, info

    def observation_spec(self):
        gym_spec = self.env.observation_space
        if isinstance(gym_spec, gym.spaces.Box):
            return self._add_modality(gym_spec.shape)
        else:
            raise ValueError('Discrete observation currently not supported')
        # TODO: migrate everything to dm_format

    def action_spec(self):
        gym_spec = self.env.action_space
        if isinstance(gym_spec, gym.spaces.Box):
            return {
                'type': 'continuous',
                'dim': gym_spec.shape
            }
        else:
            raise ValueError('Discrete observation currently not supported')

    def _close(self):
        self.env.close()

    def _render(self):
        return self.env.render(mode='rgb_array')

    @property
    def spec_format(self):
        return SpecFormat.SURREAL_CLASSIC


class RobosuiteWrapper(Wrapper):
    def __init__(self, env, env_config):
        # dm_control envs don't have metadata
        env.metadata = {}
        super().__init__(env)
        self.use_depth = env_config.use_depth and env_config.pixel_input
        self._input_list = env_config.observation
        self._action_repeat = env_config.action_repeat or 1

    def _add_modality(self, obs, verbose=False):
        pixel_modality = collections.OrderedDict()
        flat_modality = collections.OrderedDict()
        for key in obs:
            if key == 'image' and 'camera0' in self._input_list['pixel']:
                pixel_modality['camera0'] = obs[key]
            elif key in self._input_list['pixel']:
                pixel_modality[key] = obs[key]
            elif key in self._input_list['low_dim']:
                flat_modality[key] = obs[key]
            elif verbose:
                print('Mujoco: skipping observation key:', key)
        obs = collections.OrderedDict()
        if len(pixel_modality) > 0:
            obs['pixel'] = pixel_modality
        if len(flat_modality) > 0:
            obs['low_dim'] = flat_modality
        return obs

    def _step(self, action):
        rewards = []
        for repeat in range(self._action_repeat):
            obs, reward, done, info = self.env.step(action)
            rewards.append(reward)
            if done:
                break
        reward = np.mean(rewards)

        if self.use_depth:
            obs['image'] = np.concatenate((obs['image'], np.expand_dims(obs['depth'], 2)), 2)

        return self._add_modality(obs), reward, done, info

    def _reset(self):
        obs = self.env.reset()

        if self.use_depth:
            obs['image'] = np.concatenate((obs['image'], np.expand_dims(obs['depth'], 2)), 2)
        
        return self._add_modality(obs), {}

    def _close(self):
        self.env.close()

    @property
    def spec_format(self):
        return SpecFormat.MUJOCOMANIP

    def observation_spec(self):
        spec = self.env.observation_spec()

        if self.use_depth:
            spec['image'] = np.concatenate((spec['image'], np.expand_dims(spec['depth'], 2)), 2)

        for k in spec:
            spec[k] = tuple(np.array(spec[k]).shape)

        return self._add_modality(spec, verbose=True)

    def action_spec(self): # we haven't finalized the action spec of mujocomanip
        return {'dim': (self.env.dof,), 'type': 'continuous'}

    def _render(self, *args, **kwargs):
        return self.env.sim.render(camera_name='frontview',
                                   height=512,
                                   width=512,
                                   depth=False)


class ObservationConcatenationWrapper(Wrapper):
    def __init__(self, env, concatenated_obs_name='flat_inputs'):
        super().__init__(env)
        self._concatenated_obs_name = concatenated_obs_name

    def _flatten_obs(self, obs):
        flat_observations = []
        if 'low_dim' in obs:
            for modality, v in obs['low_dim'].items():
                flat_observations.append(v)
            if len(flat_observations) > 0:
                flat_observations = np.concatenate(flat_observations)
                del obs['low_dim']
                obs['low_dim'] = collections.OrderedDict([(self._concatenated_obs_name, flat_observations)])
        return obs

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        return self._flatten_obs(obs), reward, done, info

    def _reset(self):
        obs, info = self.env.reset()
        return self._flatten_obs(obs), info

    @property
    def spec_format(self):
        return SpecFormat.SURREAL_CLASSIC

    def observation_spec(self):
        spec = self.env.observation_spec()
        flat_observation_dim = 0
        if 'low_dim' in spec:
            for k, shape in spec['low_dim'].items():
                assert len(shape) == 1
                flat_observation_dim += shape[0]
            spec['low_dim'] = collections.OrderedDict([(self._concatenated_obs_name, (flat_observation_dim,))])
        return spec

    def action_spec(self):
        return self.env.action_spec()


class TransposeWrapper(Wrapper):
    def __init__(self, env):
        super().__init__(env)

    def _transpose(self, obs):
        if 'pixel' in obs:
            for key in obs['pixel']:
                # input is (H, W, 3), we want (C, H, W) == (3, 84, 84)
                obs['pixel'][key] = obs['pixel'][key].transpose((2, 0, 1))
        return obs

    def _reset(self):
        obs, info = self.env.reset()
        return self._transpose(obs), info

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        return self._transpose(obs), reward, done, info

    def observation_spec(self):
        spec = self.env.observation_spec()
        if 'pixel' in spec:
            for key in spec['pixel']:
                H, W, C = spec['pixel'][key]
                # We transpose to (C, H, W) to work with pytorch convolutions
                visual_dim = (C, H, W)
                spec['pixel'][key] = visual_dim
        return spec


class GrayscaleWrapper(Wrapper):
    def __init__(self, env):
        super().__init__(env)

    def _grayscale(self, obs):
        for key in obs['pixel']:
            observation_modality = obs['pixel'][key]
            C, H, W = observation_modality.shape
            # For now, we expect an RGB image
            assert C == 3
            obs['pixel'][key] = np.mean(observation_modality, 0, 'uint8').reshape(1, H, W)
        return obs

    def _step(self, action):
        obs, reward, done, info = self.env.step(action)
        obs = self._grayscale(obs)
        return obs, reward, done, info

    def _reset(self):
        obs, info = self.env.reset()
        scaled = self._grayscale(obs)
        return scaled, info

    @property
    def spec_format(self):
        return SpecFormat.SURREAL_CLASSIC

    def observation_spec(self):
        spec = self.env.observation_spec()
        # TODO: make constant pixel_modality='pixel'
        for key in spec['pixel']:
            dimensions = spec['pixel'][key]
            C, H, W = dimensions
            # We expect rgb for now
            assert C == 3
            spec['pixel'][key] = (1, H, W)
        return spec

    def action_spec(self):
        return self.env.action_spec()

class FrameStackWrapper(Wrapper):
    def __init__(self, env, env_config):
        super().__init__(env)
        self.n = env_config.frame_stacks
        self.frame_stack_concatenate_on_env = env_config.frame_stack_concatenate_on_env
        self._history = deque(maxlen=self.n)

    def _stacked_observation(self, obs):
        '''
        Assumes self._history contains the last n frames from the environment
        Concatenates the frames together along the depth axis
        '''
        #stacked_obs_dict_unordered = collections.OrderedDict()
        #stacked_obs_dict_unordered['pixel'] = collections.OrderedDict()
        new_pixel_modality = collections.OrderedDict()

        for key in obs['pixel']:
            obs_stacked = []
            for history_obs in self._history:
                obs_stacked.append(history_obs['pixel'][key])
            assert len(obs_stacked) <= self.n
            if self.frame_stack_concatenate_on_env:
                stacked = np.concatenate(obs_stacked, axis=0)
            else:
                stacked = obs_stacked
            #stacked_obs_dict_unordered['pixel'][key] = stacked
            new_pixel_modality[key] = stacked
        next_stacked_dict = collections.OrderedDict()
        for key in obs:
            if key == 'pixel':
                next_stacked_dict[key] = new_pixel_modality
            else:
                next_stacked_dict[key] = obs[key]
        return next_stacked_dict

    def _step(self, action):
        obs_next, reward, done, info = self.env.step(action)
        self._history.append(obs_next)
        obs_next_stacked = self._stacked_observation(obs_next)
        return obs_next_stacked, reward, done, info

    def _reset(self):
        obs, info = self.env.reset()
        for i in range(self.n):
            self._history.append(obs)
        a = self._stacked_observation(obs)
        return a, info

    @property
    def spec_format(self):
        return SpecFormat.SURREAL_CLASSIC

    def observation_spec(self):
        spec = self.env.observation_spec()
        if 'pixel' in spec:
            for key in spec['pixel']:
                dimensions = spec['pixel'][key]
                C, H, W = dimensions
                if C > 4:
                    print('WARNING: Received input of size (C, H, W) == ', dimensions)
                    print('number of channels is greater than 4')
                spec['pixel'][key] = (C * self.n, H, W)
        return spec

    def action_spec(self):
        return self.env.action_spec()

class FilterWrapper(Wrapper):
    '''
    Given the inputs allowed in env_config.observation, reject any inputs
    not specified in the config.
    '''
    def __init__(self, env, env_config):
        super().__init__(env)
        self._allowed_items = env_config.observation

    def _filtered_obs(self, obs, verbose=False):
        filtered = collections.OrderedDict()
        for modality in obs:
            if modality in self._allowed_items:
                modality_spec = collections.OrderedDict()
                for key in obs[modality]:
                    if key in self._allowed_items[modality]:
                        modality_spec[key] = obs[modality][key]
                    elif verbose:
                        print('Skipping observation key:', modality, '/', key)
                filtered[modality] = modality_spec
        return filtered

    def _step(self, action):
        obs_next, reward, done, info = self.env.step(action)
        return self._filtered_obs(obs_next), reward, done, info

    def _reset(self):
        obs, info = self.env.reset()
        return self._filtered_obs(obs), info

    @property
    def spec_format(self):
        return SpecFormat.SURREAL_CLASSIC

    def observation_spec(self):
        spec = self.env.observation_spec()
        return self._filtered_obs(spec, verbose=True)

    def action_spec(self):
        return self.env.action_spec()



import random
from collections import deque
from .base import Replay


class FIFOReplay(Replay):
    """
    WARNING: you must set session_config.replay:
    - max_puller_queue: to a very small number, like 1
    - max_prefetch_queue: to 1
    session_config.sender:
    - flush_iteration: to a small number
    """
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 index=0):
        super().__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            index=index,
        )
        self.batch_size = self.learner_config.replay.batch_size
        self.memory_size = self.learner_config.replay.memory_size
        self._memory = deque(maxlen=self.memory_size+3)  # + 3 for a gentle buffering
        assert self.session_config.replay.max_puller_queue <= 10
        assert self.session_config.replay.max_prefetch_queue == 1 
        assert not self.session_config.sender.flush_time
        assert self.session_config.sender.flush_iteration <= 10


    def insert(self, exp_tuple):
        self._memory.append(exp_tuple)

    def sample(self, batch_size):
        assert batch_size <= self.memory_size
        return [self._memory.popleft() for _ in range(batch_size)]

    def evict(self):
        raise NotImplementedError('no support for eviction in FIFO mode')

    def start_sample_condition(self):
        return len(self._memory) >= self.batch_size

    def __len__(self):
        return len(self._memory)




import os
from setuptools import setup, find_packages


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read().strip()


setup(
    name='Surreal',
    version='0.1.0',
    author='Stanford Vision and Learning Lab',
    url='http://github.com/StanfordVL/Surreal',
    description='Stanford University Repository for Reinforcement Algorithms',
    long_description=read('README.rst'),
    keywords=['Reinforcement Learning',
              'Deep Learning',
              'Distributed Computing'],
    license='GPLv3',
    packages=[
        package for package in find_packages() if package.startswith("surreal")
    ],
    entry_points={
        'console_scripts': [
            'surreal-kube=surreal.kube.surreal_kube:main',
            'surreal-tmux=surreal.tmux.surreal_tmux:main',
            'surreal-ddpg=surreal.main.ddpg_configs:main',
            'surreal-ppo=surreal.main.ppo_configs:main',
            'surreal-default-config=surreal.main.generate_default_config:main',
        ]
    },
    install_requires=[
        "gym",
        "robosuite",
        "mujoco-py<1.50.2,>=1.50.1",
        "tabulate",
        "tensorflow",
        "tensorboardX",
        "imageio",
        "pygame",
        "benedict",
        "nanolog",
        "psutil",
        "tabulate",
        "imageio",
        "caraml>=0.10.0",
        "symphony>=0.9.1",
        "torchx",
        "tensorplex",
        "cloudwise>=0.1.1",
        "Cython<0.29",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Environment :: Console",
        "Programming Language :: Python :: 3"
    ],
    include_package_data=True,
    zip_safe=False
)



import torch
import numpy as np
import torchx.nn as nnx

class RewardFilter(nnx.Module):
    """
        Keeps historical average of rewards
        Attributes:
            eps: tolerance value for computing reward filter (whitening)
                default to 10
            running_sum: running sum of all previous rewards
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
            running_sumsq: sum of square of all previous states
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
            count: number of experiences accumulated
                (Note, type is torch.cuda.FloatTensor or torch.FloatTensor)
    """
    def __init__(self, eps=1e-5):
        """
            Constructor for RewardFilter class
            Args:
                eps: tolerance value for computing reward filter (whitening)
        """
        super(RewardFilter, self).__init__()

        self.eps = eps

        # Keep some buffers for doing whitening. 
        self.register_buffer('count', torch.tensor(eps, dtype=torch.float32))
        self.register_buffer('running_sum', torch.tensor(0.0, dtype=torch.float32))
        self.register_buffer('running_sumsq', torch.tensor(0.0, dtype=torch.float32))

    def update(self, x):
        """
            Count x into historical average, updates running sums and count
            Accepts batched input
            Args:
                x: input tensor to be kept in record. 
        """
        self.count += float(np.prod(x.size()))
        self.running_sum += x.sum()
        self.running_sumsq = (x * x).sum()

    def forward(self, inputs):
        '''
            Whiten reward to have zero-mean, unit variance.
            Also clamps output to be within 5 standard deviations
            Args:
                inputs -- batched reward input. batch size at dim 0
            Returns:
                0 mean std 1 weightened batch of observation
        '''
        mean = (self.running_sum / self.count)
        std  = torch.clamp((self.running_sumsq / self.count \
                                  - mean.pow(2)).pow(0.5), min=self.eps)
        normed = torch.clamp((inputs - mean) / std, -5.0, 5.0)
        return normed

    def reward_mean(self):
        '''
            Outputs the current reward mean
        '''
        return (self.running_sum / self.count).item()



"""
Actor function
"""
import time
import torch
import numpy as np
import torchx as tx
import torchx.nn as nnx
import surreal.utils as U
from surreal.model.ppo_net import PPOModel, DiagGauss
from surreal.env import ExpSenderWrapperMultiStepMovingWindowWithInfo
from surreal.session import ConfigError
from .base import Agent


class PPOAgent(Agent):
    '''
        Class that specifies PPO agent logic
        Important attributes:
            init_log_sig: initial log sigma for diagonal gausian policy
            model: PPO_Model instance. see surreal.model.ppo_net
            pd: DiagGauss instance. see surreal.model.ppo_net
        Member functions:
            act
            reset
    '''
    def __init__(self,
                 learner_config,
                 env_config,
                 session_config,
                 agent_id,
                 agent_mode,
                 render=False):
        super().__init__(
            learner_config=learner_config,
            env_config=env_config,
            session_config=session_config,
            agent_id=agent_id,
            agent_mode=agent_mode,
            render=render,
        )
        self.action_dim = self.env_config.action_spec.dim[0]
        self.obs_spec = self.env_config.obs_spec
        self.use_z_filter = self.learner_config.algo.use_z_filter

        self.init_log_sig = self.learner_config.algo.consts.init_log_sig
        self.log_sig_range = self.learner_config.algo.consts.log_sig_range

        # setting agent mode
        if self.agent_mode != 'training':
            if self.env_config.stochastic_eval:
                self.agent_mode = 'eval_stochastic'
            else:
                self.agent_mode = 'eval_deterministic'

        if self.agent_mode != 'training':
            self.noise = 0
        else:
            self.noise = np.random.uniform(low=-self.log_sig_range,
                                           high=self.log_sig_range)
        self.rnn_config = self.learner_config.algo.rnn

        # GPU setup
        # TODO: deprecate
        self._num_gpus = session_config.agent.num_gpus

        if torch.cuda.is_available():
            self.gpu_ids = 'cuda:all'
            self.log.info('PPO agent is using GPU')
            # Note that user is responsible for only providing one GPU for the program
            self.log.info('cudnn version: {}'.format(torch.backends.cudnn.version()))
            torch.backends.cudnn.benchmark = True
        else:
            self.gpu_ids = 'cpu'
            self.log.info('PPO agent is using CPU')

        self.pd = DiagGauss(self.action_dim)
        self.cells = None

        with tx.device_scope(self.gpu_ids):
            if self.rnn_config.if_rnn_policy:
                # Note that .detach() is necessary here to prevent overflow of memory
                # otherwise rollout in length of thousands will prevent previously
                # accumulated hidden/cell states from being freed.
                self.cells = (torch.zeros(self.rnn_config.rnn_layer,
                                          1,  # batch_size is 1
                                          self.rnn_config.rnn_hidden).detach(),
                              torch.zeros(self.rnn_config.rnn_layer,
                                          1,  # batch_size is 1
                                          self.rnn_config.rnn_hidden).detach())

            self.model = PPOModel(
                obs_spec=self.obs_spec,
                action_dim=self.action_dim,
                model_config=self.learner_config.model,
                use_cuda=False,
                init_log_sig=self.init_log_sig,
                use_z_filter=self.use_z_filter,
                if_pixel_input=self.env_config.pixel_input,
                rnn_config=self.rnn_config,
            )

    def act(self, obs):
        '''
            Agent returns an action based on input observation. if in training,
            returns action along with action infos, which includes the current
            probability distribution, RNN hidden states and etc.
            Args:
                obs: numpy array of (1, obs_dim)

            Returns:
                action_choice: sampled or max likelihood action to input to env
                action_info: list of auxiliary information - [onetime, persistent]
                    Note: this includes probability distribution the action is
                    sampled from, RNN hidden states
        '''
        # Note: we collect two kinds of action infos, one persistent one onetime
        # persistent info is collected for every step in rollout (i.e. policy probability distribution)
        # onetime info is collected for the first step in partial trajectory (i.e. RNN hidden state)
        # see ExpSenderWrapperMultiStepMovingWindowWithInfo in exp_sender_wrapper for more
        action_info = [[], []]

        with tx.device_scope(self.gpu_ids):
            obs_tensor = {}
            for mod in obs.keys():
                obs_tensor[mod] = {}
                for k in obs[mod].keys():
                    obs_tensor[mod][k] = torch.tensor(obs[mod][k], dtype=torch.float32).unsqueeze(0)

            if self.rnn_config.if_rnn_policy:
                action_info[0].append(self.cells[0].squeeze(1).cpu().numpy())
                action_info[0].append(self.cells[1].squeeze(1).cpu().numpy())

            action_pd, self.cells = self.model.forward_actor_expose_cells(obs_tensor, self.cells)
            action_pd = action_pd.detach().cpu().numpy()
            action_pd[:, self.action_dim:] *= np.exp(self.noise)

            if self.agent_mode != 'eval_deterministic':
                action_choice = self.pd.sample(action_pd)
            else:
                action_choice = self.pd.maxprob(action_pd)
            np.clip(action_choice, -1, 1, out=action_choice)

            action_choice = action_choice.reshape((-1,))
            action_pd     = action_pd.reshape((-1,))
            action_info[1].append(action_pd)
            if self.agent_mode != 'training':
                return action_choice
            else:
                time.sleep(self.env_config.sleep_time)
                return action_choice, action_info

    def module_dict(self):
        return {
            'ppo': self.model,
        }

    def default_config(self):
        return {
            'model': {
                'convs': '_list_',
                'fc_hidden_sizes': '_list_',
            },
        }

    def reset(self):
        '''
            reset of LSTM hidden and cell states
        '''
        if self.rnn_config.if_rnn_policy:
            # Note that .detach() is necessary here to prevent overflow of memory
            # otherwise rollout in length of thousands will prevent previously
            # accumulated hidden/cell states from being freed.
            with tx.device_scope(self.gpu_ids):
                self.cells = (torch.zeros(self.rnn_config.rnn_layer,
                                          1,  # batch_size is 1
                                          self.rnn_config.rnn_hidden).detach(),
                              torch.zeros(self.rnn_config.rnn_layer,
                                          1,  # batch_size is 1
                                          self.rnn_config.rnn_hidden).detach())

    def prepare_env_agent(self, env):
        env = super().prepare_env_agent(env)
        env = ExpSenderWrapperMultiStepMovingWindowWithInfo(env,
                                                            self.learner_config,
                                                            self.session_config)
        return env



"""
Numpy and math operations
"""
import numpy as np


def min_at(values):
    "Returns: (min, min_i)"
    if not values:
        return None, None
    return min( (v, i) for i, v in enumerate(values) )


def max_at(values):
    "Returns: (max, max_i)"
    if not values:
        return None, None
    return max( (v, i) for i, v in enumerate(values) )


def sum_pow(p, n_begin, n_end):
    """
    Power summation, N inclusive.
    \sum_{n=N_begin}^{N_end} p^n
    """
    return (p**(n_end+1) - p**n_begin) / (p - 1.0)


def ceildiv(a, b):
    """
    Ceiling division, equivalent to math.ceil(1.0*a/b) but much faster.
    ceildiv(19, 7) == 3
    ceildiv(21, 7) == 3
    ceildiv(22, 7) == 4
    """
    return - (-a // b)


def is_div(a, b):
    " Returns: bool - does `a` divide `b`. "
    return int(a) % int(b) == 0


def cum_sum(seq):
    """
    Cumulative sum (include 0)
    """
    s = 0
    cumult = [0]
    for n in seq:
        s += n
        cumult.append(s)
    return cumult
#     return [sum(seq[:i]) for i in range(len(seq) + 1)]


def is_np_array(L, dtype=None):
    if dtype is None:
        return isinstance(L, np.ndarray)
    return isinstance(L, np.ndarray) and np.issubdtype(L.dtype, dtype)


def is_float_array(L):
    return is_np_array(L, np.float)


def is_int_array(L):
    return is_np_array(L, np.int_)


def is_np_scalar(x):
    """
    Check np types like np.int64
    """
    return isinstance(x, np.generic)


def is_np_int(x):
    return isinstance(x, np.int_)


def is_np_float(x):
    return isinstance(x, np.float_)


def np_cast(x, dtype):
    if dtype is None or x.dtype == dtype:
        return x
    else:
        return x.astype(dtype)


def compare(a, b, *, tol=1e-6):
    """
    if ||a - b|| < tol, return 0
    otherwise return float(a > b)
    """
    if abs(a - b) < tol:
        return 0.0
    elif a > b:
        return 1.0
    else:
        return -1.0


def np_clip_(x, min=None, max=None):
    """
    In-place clip. Also supports indexing
    """
    return np.clip(x, min, max, out=x)



import itertools


def setup_network(*, agents,
                  evals,
                  ps,
                  replay,
                  learner,
                  tensorplex,
                  loggerplex,
                  tensorboard):
    """
        Sets up the communication between surreal
        components using symphony

        Args:
            agents, evals (list): list of symphony processes
            ps, replay, learner, tensorplex, loggerplex, tensorboard:
                symphony processes
    """
    for proc in itertools.chain(agents, evals):
        proc.connects('ps-frontend')
        proc.connects('collector-frontend')

    ps.binds('ps-frontend')
    ps.binds('ps-backend')
    ps.connects('parameter-publish')

    replay.binds('collector-frontend')
    replay.binds('sampler-frontend')
    replay.binds('collector-backend')
    replay.binds('sampler-backend')

    learner.connects('sampler-frontend')
    learner.binds('parameter-publish')
    learner.binds('prefetch-queue')

    tensorplex.binds('tensorplex')
    loggerplex.binds('loggerplex')

    for proc in itertools.chain(agents, evals, [ps, replay, learner]):
        proc.connects('tensorplex')
        proc.connects('loggerplex')

    tensorboard.exposes({'tensorboard': 6006})



from queue import Queue
import torch.nn as nn
from torch.nn.init import xavier_uniform
import surreal.utils as U
import torch.nn.functional as F
import numpy as np
import itertools
import torchx.nn as nnx

from .model_builders import *
from .z_filter import ZFilter

class DDPGModel(nnx.Module):

    def __init__(self,
                 obs_spec,
                 action_dim,
                 use_layernorm,
                 actor_fc_hidden_sizes,
                 critic_fc_hidden_sizes,
                 conv_out_channels,
                 conv_kernel_sizes,
                 conv_strides,
                 conv_hidden_dim,
                 critic_only=False,
                 ):
        super(DDPGModel, self).__init__()

        # hyperparameters
        self.is_pixel_input = 'pixel' in obs_spec
        self.action_dim = action_dim
        self.use_layernorm = use_layernorm

        if self.is_pixel_input:
            self.input_dim = obs_spec['pixel']['camera0']
        else:
            self.input_dim = obs_spec['low_dim']['flat_inputs'][0]

        concatenated_perception_dim = 0
        if self.is_pixel_input:
            self.perception = CNNStemNetwork(self.input_dim, conv_hidden_dim, conv_channels=conv_out_channels,
                                             kernel_sizes=conv_kernel_sizes, strides=conv_strides)
            concatenated_perception_dim += conv_hidden_dim
        if 'low_dim' in obs_spec:
            concatenated_perception_dim += obs_spec['low_dim']['flat_inputs'][0]
        if not critic_only:
            self.actor = ActorNetworkX(concatenated_perception_dim, self.action_dim, hidden_sizes=actor_fc_hidden_sizes,
                                       use_layernorm=self.use_layernorm)
        else:
            self.actor = None
        self.critic = CriticNetworkX(concatenated_perception_dim, self.action_dim, hidden_sizes=critic_fc_hidden_sizes,
                                     use_layernorm=self.use_layernorm)

    def get_actor_parameters(self):
        return itertools.chain(self.actor.parameters())

    def get_critic_parameters(self):
        if self.is_pixel_input:
            return itertools.chain(self.critic.parameters(), self.perception.parameters())
        else:
            return itertools.chain(self.critic.parameters())

    def forward_actor(self, obs):
        return self.actor(obs)

    def forward_critic(self, obs, action):
        return self.critic(obs, action)

    def forward_perception(self, obs):
        concatenated_inputs = []
        if self.is_pixel_input:
            obs_pixel = obs['pixel']['camera0']
            obs_pixel = self.scale_image(obs_pixel)
            cnn_updated = self.perception(obs_pixel)
            concatenated_inputs.append(cnn_updated)
        if 'low_dim' in obs:
            concatenated_inputs.append(obs['low_dim']['flat_inputs'])
        concatenated_inputs = torch.cat(concatenated_inputs, dim=1)
        return concatenated_inputs

    def forward(self, obs_in, calculate_value=True, action=None):
        obs_in = self.forward_perception(obs_in)
        if action is None:
            action = self.forward_actor(obs_in)
        value = None
        if calculate_value:
            value = self.forward_critic(obs_in, action)
        return action, value

    def scale_image(self, obs, scaling_factor=255.0):
        '''
        Given uint8 input from the environment, scale to float32 and
        divide by 255 to scale inputs between 0.0 and 1.0
        '''
        return obs / scaling_factor
        


from surreal.launch.setup_network import setup_network


def create_surreal_containerized(exp,
                                 nonagent_image,
                                 agent_image,
                                 cmd_dict,
                                 batched=False):
    """
    TODO: document

    Args:
        exp: [description]
        nonagent_image: [description]
        agent_image: [description]
        cmd_dict: [description]
        batched: [description] (default: {False})
    """
    nonagent = exp.new_process_group('nonagent')
    learner = nonagent.new_process(
        'learner',
        container_image=nonagent_image,
        args=[cmd_dict['learner']])

    # For dm_control
    learner.set_env('DISABLE_MUJOCO_RENDERING', "1")

    replay = nonagent.new_process(
        'replay',
        container_image=nonagent_image,
        args=[cmd_dict['replay']])

    ps = nonagent.new_process(
        'ps',
        container_image=nonagent_image,
        args=[cmd_dict['ps']])

    tensorboard = nonagent.new_process(
        'tensorboard',
        container_image=nonagent_image,
        args=[cmd_dict['tensorboard']])

    tensorplex = nonagent.new_process(
        'tensorplex',
        container_image=nonagent_image,
        args=[cmd_dict['tensorplex']])

    loggerplex = nonagent.new_process(
        'loggerplex',
        container_image=nonagent_image,
        args=[cmd_dict['loggerplex']])
    nonagent.image_pull_policy('Always')

    agents = []
    for i, arg in enumerate(cmd_dict['agent']):
        agent_fmt = 'agent-{}'
        if batched:
            agent_fmt = 'agents-{}'
        agent = exp.new_process(
            agent_fmt.format(i),
            container_image=agent_image,
            args=[arg])

        agents.append(agent)

    evals = []
    for i, arg in enumerate(cmd_dict['eval']):
        eval_fmt = 'eval-{}'
        if batched:
            eval_fmt = 'evals-{}'
        eval_p = exp.new_process(
            eval_fmt.format(i),
            container_image=agent_image,
            args=[arg])

        evals.append(eval_p)

    setup_network(agents=agents,
                  evals=evals,
                  learner=learner,
                  replay=replay,
                  ps=ps,
                  tensorboard=tensorboard,
                  tensorplex=tensorplex,
                  loggerplex=loggerplex)
    return {
        'agents': agents,
        'evals': evals,
        'learner': learner,
        'replay': replay,
        'ps': ps,
        'tensorboard': tensorboard,
        'tensorplex': tensorplex,
        'loggerplex': loggerplex
    }



import os
import sys
import math
from copy import copy
from symphony.commandline import SymphonyParser
from symphony.engine import SymphonyConfig, Cluster
from benedict import BeneDict
from surreal.launch import (
    CommandGenerator,
    setup_network,
)
import surreal.utils as U


def _merge_setting_dictionaries(customize, base):
    di = copy(base)
    for key in di:
        if isinstance(di[key], dict):
            if key in customize:
                di[key] = _merge_setting_dictionaries(customize[key], di[key])
        else:
            if key in customize and customize[key] is not None:
                di[key] = customize[key]
    return di


class surreal-tmuxParser(SymphonyParser):
    def create_cluster(self):
        return Cluster.new('tmux')

    def setup(self):
        super().setup()
        self.config = BeneDict()
        self.load_config()
        self._setup_create()

    def load_config(self):
        surreal_yml_path = U.get_config_file()
        if not U.f_exists(surreal_yml_path):
            raise ValueError('Cannot find surreal config file at {}'
                             .format(surreal_yml_path))
        self.config = BeneDict.load_yaml_file(surreal_yml_path)
        SymphonyConfig().set_username(self.username)
        SymphonyConfig().set_experiment_folder(self.folder)

    @property
    def folder(self):
        if 'tmux_results_folder' not in self.config:
            raise KeyError('Please specify "tmux_results_folder" in ~/.surreal.yml')
        return U.f_expand(self.config.tmux_results_folder)

    @property
    def username(self):
        if 'username' not in self.config:
            raise KeyError('Please specify "username" in ~/.surreal.yml')
        return self.config.username

    def _setup_create(self):
        parser = self.add_subparser('create', aliases=['c'])
        self._add_experiment_name(parser)
        parser.add_argument(
            '--algorithm',
            type=str,
            default='ppo',
            help='ddpg / ppo or the location of algorithm python script'
        )
        parser.add_argument(
            '--num_agents',
            type=int,
            default=2,
            help='number of agent pods to run in parallel.'
        )
        parser.add_argument(
            '--num_evals',
            type=int,
            default=1,
            help='number of eval pods to run in parallel.'
        )
        parser.add_argument(
            '--env',
            type=str,
            default='gym:HalfCheetah-v2',
            help='What environment to run'
        )
        parser.add_argument(
            '--gpu',
            type=str,
            default="auto",
            help='Which gpus to use: "auto" (default) uses all gpus '
            'available through CUDA_VISIBLE_DEVICES, or use a '
            'comma seperated list to override'
        )
        self._add_dry_run(parser)

    # ==================== helpers ====================
    def _add_dry_run(self, parser):
        parser.add_argument(
            '-dr', '--dry-run',
            action='store_true',
            help='print the kubectl command without actually executing it.'
        )

    def _process_experiment_name(self, experiment_name):
        """
        experiment_name will be used as DNS, so must not have underscore or dot
        """
        new_name = experiment_name.lower().replace('.', '-').replace('_', '-')
        if new_name != experiment_name:
            print('experiment name string has been fixed: {} -> {}'
                  .format(experiment_name, new_name))
        return new_name

    def action_create(self, args):
        """
            Spin up a multi-node distributed Surreal experiment.
            Put any command line args that pass to the config script after "--"
        """
        cluster = self.create_cluster()
        experiment_name = self._process_experiment_name(args.experiment_name)
        exp = cluster.new_experiment(
            experiment_name,
            preamble_cmds=self.config.tmux_preamble_cmds)

        algorithm_args = args.remainder
        algorithm_args += [
            "--num-agents",
            str(args.num_agents),
            ]
        experiment_folder = os.path.join(self.folder, experiment_name)
        print('Writing experiment output to {}'.format(experiment_folder))
        algorithm_args += ["--experiment-folder",
                           experiment_folder]
        algorithm_args += ["--env", args.env]
        executable = self._find_executable(args.algorithm)
        cmd_gen = CommandGenerator(
            num_agents=args.num_agents,
            num_evals=args.num_evals,
            executable=executable,
            config_commands=algorithm_args)

        learner = exp.new_process(
            'learner',
            cmds=[cmd_gen.get_command('learner')])
        # learner.set_env('DISABLE_MUJOCO_RENDERING', "1")

        replay = exp.new_process(
            'replay',
            cmds=[cmd_gen.get_command('replay')])

        ps = exp.new_process(
            'ps',
            cmds=[cmd_gen.get_command('ps')])

        tensorboard = exp.new_process(
            'tensorboard',
            cmds=[cmd_gen.get_command('tensorboard')])

        tensorplex = exp.new_process(
            'tensorplex',
            cmds=[cmd_gen.get_command('tensorplex')])

        loggerplex = exp.new_process(
            'loggerplex',
            cmds=[cmd_gen.get_command('loggerplex')])

        agents = []
        for i in range(args.num_agents):
            agent_name = 'agent-{}'.format(i)
            agent = exp.new_process(
                agent_name,
                cmds=[cmd_gen.get_command(agent_name)])
            agents.append(agent)

        evals = []
        for i in range(args.num_evals):
            eval_name = 'eval-{}'.format(i)
            eval_p = exp.new_process(
                eval_name,
                cmds=[cmd_gen.get_command(eval_name)])
            evals.append(eval_p)

        setup_network(agents=agents,
                      evals=evals,
                      learner=learner,
                      replay=replay,
                      ps=ps,
                      tensorboard=tensorboard,
                      tensorplex=tensorplex,
                      loggerplex=loggerplex)
        self._setup_gpu(agents=agents,
                        evals=evals,
                        learner=learner,
                        gpus=args.gpu)
        cluster.launch(exp, dry_run=args.dry_run)

    def _find_executable(self, name):
        """
            Finds the .py file corresponding to the algorithm specified

        Args:
            name: ddpg / ppo / <path in container to compatible .py file>
        """
        if name == 'ddpg':
            return 'surreal-ddpg'
        elif name == 'ppo':
            return 'surreal-ppo'
        else:
            return name

    def _setup_gpu(self, agents, evals, learner, gpus):
        """
            Assigns GPU to agents and learners in an optimal way.
            No GPU, do nothing
        """
        actors = agents + evals
        if gpus == "auto":
            if 'CUDA_VISIBLE_DEVICES' in os.environ:
                gpus = os.environ['CUDA_VISIBLE_DEVICES']
            else:
                gpus = ''
        gpus_str = gpus
        try:
            gpus = [x for x in gpus_str.split(',') if len(x) > 0]
        except Exception as e:
            print("Error parsing GPU specification {}\n".format(gpus_str),
                  file=sys.stderr)
            raise e
        if len(gpus) == 0:
            print('Using CPU')
        elif len(gpus) == 1:
            gpu = gpus[0]
            print('Putting agents, evals and learner on GPU {}'.format(gpu))
            for actor in actors + [learner]:
                actor.set_envs({'CUDA_VISIBLE_DEVICES': gpu})
        elif len(gpus) > 1:
            learner_gpu = gpus[0]
            print('Putting learner on GPU {}'.format(learner_gpu))
            learner.set_envs({'CUDA_VISIBLE_DEVICES': learner_gpu})

            actors_per_gpu = float(len(actors)) / (len(gpus) - 1)
            actors_per_gpu = int(math.ceil(actors_per_gpu))
            print('Putting up to {} agents/evals on each of gpus {}'.format(
                    actors_per_gpu, ','.join([x for x in gpus[1:]])))

            for i, actor in enumerate(actors):
                cur_gpu = gpus[1 + i // actors_per_gpu]
                actor.set_envs({'CUDA_VISIBLE_DEVICES': cur_gpu})


def main():
    surreal-tmuxParser().main()


if __name__ == '__main__':
    main()



from .base import *
from .atari_names import *
from .atari_wrappers import *
from .exp_sender_wrapper import *
from .monitor import *
from .wrapper import *
from .make_env import make_env
from .video_env import VideoWrapper